/* doc.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023, 2024 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdbool.h>

#include "doc.h"

Image * history_point_detach_bgimg(HistoryPoint *this)
{
	Image *ret = this->bgimg;
	this->bgimg = NULL;

	return ret;
}

Image * history_point_detach_fgimg(HistoryPoint *this)
{
	Image *ret = this->fgimg;
	this->fgimg = NULL;

	return ret;
}

void history_point_construct(HistoryPoint *this)
{
	this->bgimg = NULL;
	this->fgimg = NULL;
}

void history_point_destruct(HistoryPoint *this)
{
	if(this->fgimg) {
		image_destruct((Image *) this->fgimg);
		free(this->fgimg);
	}

	if(this->bgimg) {
		image_destruct((Image *) this->bgimg);
		free(this->bgimg);
	}

}

void history_point_stack_construct(HistoryPointStack *this)
{
	this->count = 0;
	HistoryPoint * *_tmp_2 = (HistoryPoint * *) malloc(8 * sizeof(HistoryPoint));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	this->arr = _tmp_2;
	this->alloccount = 8;
}

_Bool history_point_stack_is_empty(HistoryPointStack *this)
{
	return 0 == this->count;
}

void history_point_stack_push(HistoryPointStack *this, HistoryPoint *newitem)
{
	history_point_stack__resize(this, this->count + 1);
	this->arr[this->count - 1] = newitem;
}

void history_point_stack_push_multiple(HistoryPointStack *this, int n, HistoryPoint * *newitems)
{
	int i;
	int newcount = this->count + n;
	if(this->alloccount < newcount) {
		this->arr = realloc(this->arr, newcount * sizeof(this->arr[0]));
		if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

		this->alloccount = newcount;
	}

	for(i = 0; i < n; i += 1) {
		this->arr[this->count + i] = newitems[i];
	}

	this->count = newcount;
}

HistoryPoint * history_point_stack_peek(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:42 */

	return this->arr[this->count - 1];
}

HistoryPoint * history_point_stack_pop(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:48 */
	HistoryPoint *retitem = this->arr[this->count - 1];
	history_point_stack__resize(this, this->count - 1);

	return retitem;
}

HistoryPoint * history_point_stack_pop_bottom(HistoryPointStack *this)
{
	assert(this->count > 0); /* stack.ngg:61 */
	HistoryPoint *retval = this->arr[0];
	this->count -= 1;
	memmove(this->arr, this->arr + 1, this->count * sizeof(this->arr[0]));

	return retval;
}

void history_point_stack_replace_top(HistoryPointStack *this, HistoryPoint *newitem)
{
	assert(this->count > 0); /* stack.ngg:73 */
	this->arr[this->count - 1] = newitem;
}

void history_point_stack__resize(HistoryPointStack *this, int newcount)
{
	int UNIT = 8;
	int newalloccount = this->alloccount;
	if(this->alloccount != newcount) {
		if(this->alloccount < newcount) {
			newalloccount = this->alloccount + UNIT;
		} else {
			int spare = this->alloccount - newcount;
			if(spare > (2 * UNIT)) {
				newalloccount = this->alloccount - UNIT;
			}

		}

		assert(newalloccount >= this->count); /* stack.ngg:104 */
		if(newalloccount != this->alloccount) {
			if(newalloccount == 0) {
			} else {
				this->arr = realloc(this->arr, newalloccount * sizeof(this->arr[0]));
				if(this->arr == NULL) { perror(NULL); exit(EXIT_FAILURE); }

			}

			this->alloccount = newalloccount;
		}

	}

	this->count = newcount;
}

void history_point_stack_destruct(HistoryPointStack *this)
{
	if(this->arr) {
		free(this->arr);
	}

}

struct DrawingParams default_drawparams()
{
	struct DrawingParams drawparams = drawing_params_default();
	drawparams.fgcolor = PIXEL_BLACK;
	drawparams.brushparams.opacity = 1;
	drawparams.brushparams.siz = 2;

	return drawparams;
}

void image_document_construct(ImageDocument *this, int width, int height)
{
	HistoryPointStack *_tmp_2 = (HistoryPointStack *) malloc(sizeof(HistoryPointStack));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_stack_construct(_tmp_2);
	this->redostack = _tmp_2;
	HistoryPointStack *_tmp_4 = (HistoryPointStack *) malloc(sizeof(HistoryPointStack));
	if(_tmp_4 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_stack_construct(_tmp_4);
	this->undostack = _tmp_4;
	this->autocenter = false;
	this->oldpreviewzoom = 0.0;
	this->oldpreviewh = 0;
	this->oldprevieww = 0;
	this->oldpreviewy = 0;
	this->oldpreviewx = 0;
	this->vpy = 0;
	this->vpx = 0;
	this->oldy = 0.0;
	this->oldx = 0.0;
	this->pen_is_down = false;
	this->preview = NULL;
	this->dirtyrect = rectangle_default();
	assert(width > 0); /* doc.ngg:97 */
	assert(height > 0); /* doc.ngg:98 */
	this->bgcolor = PIXEL_WHITE;
	Image *_tmp_5 = (Image *) malloc(sizeof(Image));
	if(_tmp_5 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_5, width, height, 255);
	this->checkpattern = _tmp_5;
	image_fill_checkpattern(this->checkpattern);
	this->drawparams = default_drawparams();
	this->stroke_smoothness = 0.1;
	Layer *_tmp_6 = (Layer *) malloc(sizeof(Layer));
	if(_tmp_6 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	layer_construct(_tmp_6, width, height, PIXEL_WHITE);
	this->bglayer = _tmp_6;
	Layer *_tmp_7 = (Layer *) malloc(sizeof(Layer));
	if(_tmp_7 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	layer_construct(_tmp_7, width, height, PIXEL_TRANSP);
	this->fglayer = _tmp_7;
	this->layer = this->fglayer;
	this->fgalpha = 1;
	this->bgvisible = true;
	this->subpixel_when_zoomed = true;
	this->vpzoom = 1;
	Image *_tmp_8 = (Image *) malloc(sizeof(Image));
	if(_tmp_8 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_8, width, height, 255);
	this->composite = _tmp_8;
	image_document_mark_all_dirty(this);
}

void image_document_destruct(ImageDocument *this)
{
	image_document_clear_undostack(this);
	image_document_clear_redostack(this);
	if(this->checkpattern) {
		image_destruct(this->checkpattern);
		free(this->checkpattern);
	}

	if(this->fglayer) {
		layer_destruct(this->fglayer);
		free(this->fglayer);
	}

	if(this->bglayer) {
		layer_destruct(this->bglayer);
		free(this->bglayer);
	}

	if(this->composite) {
		image_destruct(this->composite);
		free(this->composite);
	}

	if(this->preview) {
		image_destruct(this->preview);
		free(this->preview);
	}

	if(this->undostack) {
		history_point_stack_destruct(this->undostack);
		free(this->undostack);
	}

	if(this->redostack) {
		history_point_stack_destruct(this->redostack);
		free(this->redostack);
	}

}

void image_document_apply_snapshot(ImageDocument *this, HistoryPoint *hp)
{
	if(hp->fgimg) {
		assert(!hp->bgimg); /* doc.ngg:132 */
		layer_replace_image(this->fglayer, history_point_detach_fgimg(hp));
	} else {
		assert(!hp->fgimg); /* doc.ngg:136 */
		layer_replace_image(this->bglayer, history_point_detach_bgimg(hp));
	}

	image_document_mark_all_dirty(this);
}

_Bool image_document_can_redo(ImageDocument *this)
{
	return !history_point_stack_is_empty(this->redostack);
}

_Bool image_document_can_undo(ImageDocument *this)
{
	return !history_point_stack_is_empty(this->undostack);
}

void image_document_clear_layer(ImageDocument *this)
{
	image_document_fill_layer(this, PIXEL_TRANSP);
}

void image_document_clear_redostack(ImageDocument *this)
{
	while(!history_point_stack_is_empty(this->redostack)) {
		HistoryPoint *todel = history_point_stack_pop(this->redostack);
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

}

void image_document_clear_undostack(ImageDocument *this)
{
	while(!history_point_stack_is_empty(this->undostack)) {
		HistoryPoint *todel = history_point_stack_pop(this->undostack);
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

}

HistoryPoint * image_document_create_history_point(ImageDocument *this)
{
	HistoryPoint *hp = (HistoryPoint *) malloc(sizeof(HistoryPoint));
	if(hp == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	history_point_construct(hp);
	if(this->layer == this->bglayer) {
		hp->bgimg = image_Clone(this->bglayer->img);
	} else {
		hp->fgimg = image_Clone(this->fglayer->img);
	}


	return hp;
}

void image_document_create_redo_snapshot(ImageDocument *this)
{
	history_point_stack_push(this->redostack, image_document_create_history_point(this));
}

void image_document_create_undo_snapshot(ImageDocument *this)
{
	image_document_clear_redostack(this);
	if(this->undostack->count > STACKLIMIT) {
		HistoryPoint *todel = history_point_stack_pop_bottom(this->undostack);
		assert(this->undostack->count == STACKLIMIT); /* doc.ngg:188 */
		if(todel) {
			history_point_destruct(todel);
			free(todel);
		}

	}

	history_point_stack_push(this->undostack, image_document_create_history_point(this));
}

void image_document_dip_eyedropper(ImageDocument *this, int x, int y)
{
	assert(this->drawparams.active_tool == TOOL_EYEDROPPER); /* doc.ngg:195 */
	this->drawparams.fgcolor = image_document_get_color_at(this, x, y);
}

void image_document_export_document(ImageDocument *this, const char * filename, _Bool fillbg)
{
	image_document_mark_all_dirty(this);
	image_document_update_composite(this, false, fillbg);
	write_image_to_file(this->composite, filename);
}

void image_document_export_layer(ImageDocument *this, const char * filename, _Bool fillbg)
{
	if(this->layer == this->bglayer) {
		write_image_to_file(this->bglayer->img, filename);
	} else if(!fillbg) {
		write_image_to_file(this->fglayer->img, filename);
	} else {
		Image *comp = (Image *) malloc(sizeof(Image));
		if(comp == NULL) {
			perror(NULL);
			exit(EXIT_FAILURE);
		}

		image_construct(comp, this->fglayer->img->width, this->fglayer->img->height, this->fglayer->img->depth);
		image_fill(comp, this->bgcolor);
		image_overlay_image_with_alpha(comp, this->fglayer->img, this->fgalpha);
		write_image_to_file(comp, filename);
		if(comp) {
			image_destruct(comp);
			free(comp);
		}

	}

}

void image_document_fill_layer(ImageDocument *this, struct Pixel color)
{
	image_document_create_undo_snapshot(this);
	image_fill(this->layer->img, color);
	image_document_mark_all_dirty(this);
}

void image_document_fill_layer_with_bgcolor(ImageDocument *this)
{
	image_document_fill_layer(this, this->layer->default_bgcolor);
}

void image_document_fill_layer_with_fgcolor(ImageDocument *this)
{
	image_document_fill_layer(this, this->drawparams.fgcolor);
}

void image_document_floodfill(ImageDocument *this, int x, int y)
{
	assert(this->drawparams.active_tool == TOOL_FLOODFILL); /* doc.ngg:246 */
	double tx = image_document_translate_x(this, x);
	double ty = image_document_translate_y(this, y);
	if(image_is_point_outside(this->layer->img, (int) tx, (int) ty)) {
		return ;
	}

	image_document_create_undo_snapshot(this);
	struct Rectangle rect = nan_img_floodfill(this->layer->img, (int) tx, (int) ty, this->drawparams);
	image_document_add_dirty_region(this, rect);
}

struct Pixel image_document_get_color_at(ImageDocument *this, int x, int y)
{
	double tx = image_document_translate_x(this, x);
	double ty = image_document_translate_y(this, y);
	if((tx >= 0) && (tx < this->layer->img->width) && (ty >= 0) && (ty < this->layer->img->height)) {
		struct Pixel px = nan_img_get_pixel(this->layer->img, (int) tx, (int) ty);
		px.rgb[3] = 1;

		return px;
	}


	return PIXEL_VPBG;
}

int image_document_get_height(ImageDocument *this)
{
	return this->fglayer->img->height;
}

int image_document_get_width(ImageDocument *this)
{
	return this->fglayer->img->width;
}

void image_document_add_dirty_region(ImageDocument *this, struct Rectangle rect)
{
	if(region_is_zero(this->dirtyrect)) {
		this->dirtyrect = rect;

		return ;
	}

	int x2 = imax2(rect.x + rect.w, this->dirtyrect.x + this->dirtyrect.w);
	int y2 = imax2(rect.y + rect.h, this->dirtyrect.y + this->dirtyrect.h);
	if(rect.x < this->dirtyrect.x) {
		this->dirtyrect.x = rect.x;
	}

	if(rect.y < this->dirtyrect.y) {
		this->dirtyrect.y = rect.y;
	}

	this->dirtyrect.w = x2 - this->dirtyrect.x;
	this->dirtyrect.h = y2 - this->dirtyrect.y;
}

_Bool image_document_is_all_dirty(ImageDocument *this)
{
	return (this->dirtyrect.x == 0) && (this->dirtyrect.y == 0) && (this->dirtyrect.w == this->composite->width) && (this->dirtyrect.h == this->composite->height);
}

_Bool image_document_is_dirty(ImageDocument *this)
{
	return !region_is_zero(this->dirtyrect);
}

void image_document_mark_all_dirty(ImageDocument *this)
{
	this->dirtyrect.x = 0;
	this->dirtyrect.y = 0;
	this->dirtyrect.w = this->composite->width;
	this->dirtyrect.h = this->composite->height;
}

void image_document_mark_not_dirty(ImageDocument *this)
{
	this->dirtyrect.w = 0;
	this->dirtyrect.h = 0;
}

void image_document_pen_down(ImageDocument *this, int x, int y, double pressure)
{
	image_document_create_undo_snapshot(this);
	this->pen_is_down = true;
	this->oldx = x;
	this->oldy = y;
	image_document_pen_stroke(this, x, y, pressure);
}

void image_document_pen_up(ImageDocument *this, int x, int y, double pressure)
{
	image_document_pen_stroke(this, x, y, pressure);
	this->pen_is_down = false;
}

void image_document_pen_stroke(ImageDocument *this, int ix, int iy, double pressure)
{
	image_document_update_pressure(this, pressure);
	double x = ix;
	double y = iy;
	if(this->stroke_smoothness > 0) {
		double dist = this->stroke_smoothness * 20;
		x = this->oldx + ((x - this->oldx) / dist);
		y = this->oldy + ((y - this->oldy) / dist);
	}

	double tx1 = image_document_translate_x(this, this->oldx);
	double ty1 = image_document_translate_y(this, this->oldy);
	double tx2 = image_document_translate_x(this, x);
	double ty2 = image_document_translate_y(this, y);
	if((!this->subpixel_when_zoomed) && (this->stroke_smoothness <= 0)) {
		tx1 = (int) tx1;
		ty1 = (int) ty1;
		tx2 = (int) tx2;
		ty2 = (int) ty2;
	}

	struct Rectangle rect = image_line(this->layer->img, tx1, ty1, tx2, ty2, this->drawparams);
	image_document_add_dirty_region(this, rect);
	this->oldx = x;
	this->oldy = y;
}

double image_document_translate_x(ImageDocument *this, double x)
{
	assert(this->vpzoom > 0); /* doc.ngg:378 */

	return (x - this->vpx) / this->vpzoom;
}

double image_document_translate_y(ImageDocument *this, double y)
{
	assert(this->vpzoom > 0); /* doc.ngg:383 */

	return (y - this->vpy) / this->vpzoom;
}

void image_document_update_pressure(ImageDocument *this, double pressure)
{
	this->drawparams.oldpressure = this->drawparams.pressure;
	this->drawparams.pressure = pressure;
}

void image_document_undo(ImageDocument *this)
{
	history_point_stack_push(this->redostack, image_document_create_history_point(this));
	HistoryPoint *_ngg_tmp_1 = history_point_stack_pop(this->undostack);
	image_document_apply_snapshot(this, _ngg_tmp_1);
	if(_ngg_tmp_1) {
		history_point_destruct(_ngg_tmp_1);
		free(_ngg_tmp_1);
	}

}

void image_document_viewport_align_center(ImageDocument *this)
{
	assert(this->preview); /* doc.ngg:398 */
	this->vpx = (int) ((this->preview->width / 2.0) - ((this->vpzoom * this->composite->width) / 2.0));
	this->vpy = (int) ((this->preview->height / 2.0) - ((this->vpzoom * this->composite->height) / 2.0));
	this->autocenter = true;
}

void image_document_viewport_set_offsetx(ImageDocument *this, int x)
{
	this->vpx = x;
	this->autocenter = false;
}

void image_document_viewport_set_offsety(ImageDocument *this, int y)
{
	this->vpy = y;
	this->autocenter = false;
}

void image_document_viewport_set_size(ImageDocument *this, int w, int h)
{
	if(this->preview && (w == this->preview->width) && (h == this->preview->height)) {
		return ;
	}

	if(this->preview) {
		image_destruct(this->preview);
		free(this->preview);
		this->preview = NULL;
	}

	Image *_tmp_2 = (Image *) malloc(sizeof(Image));
	if(_tmp_2 == NULL) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	image_construct(_tmp_2, w, h, 255);
	this->preview = _tmp_2;
	if(this->autocenter) {
		image_document_viewport_align_center(this);
	}

}

void image_document_viewport_zoom_at(ImageDocument *this, int x, int y, double zoom)
{
	image_document_viewport_zoom_at_internal(this, x, y, zoom);
	this->autocenter = false;
}

void image_document_viewport_zoom_at_internal(ImageDocument *this, int x, int y, double zoom)
{
	double srcx = image_document_translate_x(this, x);
	double srcy = image_document_translate_y(this, y);
	this->vpzoom = zoom;
	this->vpx = (int) (x - (this->vpzoom * srcx));
	this->vpy = (int) (y - (this->vpzoom * srcy));
}

void image_document_viewport_zoom_at_center(ImageDocument *this, double zoom)
{
	assert(this->preview); /* doc.ngg:446 */
	image_document_viewport_zoom_at_internal(this, this->preview->width / 2, this->preview->height / 2, zoom);
}

_ngg_tuple_ImageDocument__update_composite image_document_update_composite(ImageDocument *this, _Bool fillchecker, _Bool fillbg)
{
	if(!image_document_is_dirty(this)) {
		struct Rectangle defrect = rectangle_default();

		return (_ngg_tuple_ImageDocument__update_composite){false, defrect};
	}

	assert(!(fillchecker && fillbg)); /* doc.ngg:460 */
	if(fillchecker) {
		image_copy_image_region(this->composite, this->checkpattern, this->dirtyrect);
		if(this->bgvisible) {
			image_overlay_image_region(this->composite, this->bglayer->img, this->dirtyrect);
		}

	} else if(fillbg) {
		image_fill_region(this->composite, this->bgcolor, this->dirtyrect);
		if(this->bgvisible) {
			image_overlay_image_region(this->composite, this->bglayer->img, this->dirtyrect);
		}

	} else {
		if(this->bgvisible) {
			if(image_document_is_all_dirty(this)) {
				image_copy_image(this->composite, this->bglayer->img);
			} else {
				image_copy_image_region(this->composite, this->bglayer->img, this->dirtyrect);
			}

		}

	}

	if(image_document_is_all_dirty(this)) {
		image_overlay_image_with_alpha(this->composite, this->fglayer->img, this->fgalpha);
	} else {
		image_overlay_image_region_with_alpha(this->composite, this->fglayer->img, this->fgalpha, this->dirtyrect);
	}

	if(DEBUG_COMPOSITE_DIRTYRECT) {
		print_region(this->dirtyrect, "composite dirtyrect");
		image_draw_debug_region(this->composite, this->dirtyrect);
	}

	struct Rectangle prvdirtyrect = this->dirtyrect;
	image_document_mark_not_dirty(this);

	return (_ngg_tuple_ImageDocument__update_composite){true, prvdirtyrect};
}

struct Rectangle image_document_update_preview(ImageDocument *this)
{
	assert(this->preview); /* doc.ngg:508 */
	_ngg_tuple_ImageDocument__update_composite _ngg_tmp_0 = image_document_update_composite(this, true, false);
	struct Rectangle compdirtyrect = _ngg_tmp_0.m1;
	_Bool isdirty = _ngg_tmp_0.m0;
	if(isdirty) {
		if(REPREVIEW_DIRTY_ONLY && SCALING_AA_ENABLE) {
			double border = this->vpzoom * SCALING_AA_RADIUS;
			compdirtyrect.x -= border;
			compdirtyrect.y -= border;
			compdirtyrect.w += (2 * border);
			compdirtyrect.h += (2 * border);
			compdirtyrect = image_clip_region(this->composite, compdirtyrect);
		}

	}

	_Bool geometry_changed = ((((this->oldpreviewx != this->vpx) || (this->oldpreviewy != this->vpy)) || (this->oldprevieww != this->preview->width)) || (this->oldpreviewh != this->preview->height)) || (this->oldpreviewzoom != this->vpzoom);
	if(geometry_changed) {
		image_fill(this->preview, PIXEL_VPBG);
	}

	_Bool redraw_all = (!REPREVIEW_DIRTY_ONLY) || geometry_changed;
	struct Rectangle prevdirtyrect = rectangle_default();
	if(this->vpzoom == 1) {
		if(redraw_all) {
			prevdirtyrect = image_copy_image_with_destoffset(this->preview, this->composite, this->vpx, this->vpy);
		} else if(isdirty) {
			prevdirtyrect = image_copy_image_region_with_destoffset(this->preview, this->composite, compdirtyrect, this->vpx, this->vpy);
		}

	} else {
		if(redraw_all) {
			prevdirtyrect = image_copy_image_with_scale(this->preview, this->composite, this->vpx, this->vpy, this->vpzoom);
		} else if(isdirty) {
			prevdirtyrect = image_copy_image_region_with_scale(this->preview, this->composite, compdirtyrect, this->vpx, this->vpy, this->vpzoom);
		}

	}

	this->oldpreviewx = this->vpx;
	this->oldpreviewy = this->vpy;
	this->oldprevieww = this->preview->width;
	this->oldpreviewh = this->preview->height;
	this->oldpreviewzoom = this->vpzoom;
	if(geometry_changed) {
		return image_get_region(this->preview);
	}


	return prevdirtyrect;
}

void image_document_redo(ImageDocument *this)
{
	history_point_stack_push(this->undostack, image_document_create_history_point(this));
	HistoryPoint *_ngg_tmp_2 = history_point_stack_pop(this->redostack);
	image_document_apply_snapshot(this, _ngg_tmp_2);
	if(_ngg_tmp_2) {
		history_point_destruct(_ngg_tmp_2);
		free(_ngg_tmp_2);
	}

}

void image_document_set_bgvisible(ImageDocument *this, _Bool visible)
{
	this->bgvisible = visible;
	image_document_mark_all_dirty(this);
}

void image_document_set_fgalpha(ImageDocument *this, double alpha)
{
	this->fgalpha = alpha;
	image_document_mark_all_dirty(this);
}

void image_fill_checkpattern(Image *img)
{
	int y;
	for(y = 0; y < img->height; y += 1) {
		int x;
		for(x = 0; x < img->width; x += 1) {
			_Bool islight = 0 != (((x >> 3) & 1) ^ ((y >> 3) & 1));
			nan_img_set_pixel(img, x, y, (islight ? PIXEL_CHECKER_LIGHT : PIXEL_CHECKER_DARK));
		}

	}

}

int write_image_to_file(Image *img, const char * fname)
{
	return write_image_to_png(img, fname);
}

_ngg_tuple_ImageDocument__update_composite _ngg_tuple__image_document__update_composite_default()
{
	_ngg_tuple_ImageDocument__update_composite s;
	s.m0 = false;
	s.m1 = rectangle_default();

	return s;
}

